/**
 * Copyright (c) 2012 - 2018 Data In Motion and others.
 * All rights reserved. 
 * 
 * This program and the accompanying materials are made available under the terms of the 
 * Eclipse Public License v1.0 which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors:
 *     Data In Motion - initial API and implementation
 */
package org.gecko.rsa.model;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.fail;

import java.io.IOException;
import java.io.InputStream;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.resource.ResourceSet;
import org.eclipse.emf.ecore.resource.impl.ResourceSetImpl;
import org.eclipse.emf.ecore.xml.type.XMLTypePackage;
import org.gecko.rsa.model.rsa.Array;
import org.gecko.rsa.model.rsa.DocumentRoot;
import org.gecko.rsa.model.rsa.EndpointDescription;
import org.gecko.rsa.model.rsa.EndpointDescriptions;
import org.gecko.rsa.model.rsa.Property;
import org.gecko.rsa.model.rsa.RSAFactory;
import org.gecko.rsa.model.rsa.RSAPackage;
import org.gecko.rsa.model.rsa.Value;
import org.gecko.rsa.model.rsa.configuration.RSAConfigurationComponent;
import org.gecko.rsa.model.rsa.configuration.RSAEPackageConfigurator;
import org.gecko.rsa.model.rsa.util.RSAResourceFactoryImpl;
import org.junit.Test;

public class RSAEndpointDescriptionTest {

	@Test
	public void testReadDescription() throws IOException {
		ResourceSet rs = new ResourceSetImpl();
		rs.getPackageRegistry().put(RSAPackage.eNS_URI, RSAPackage.eINSTANCE);
		rs.getResourceFactoryRegistry().getExtensionToFactoryMap().put("xml", new RSAResourceFactoryImpl());
		InputStream is = getClass().getResource("ed1.xml").openStream();
		Resource resource = rs.createResource(URI.createURI("ed.xml"), RSAPackage.eCONTENT_TYPE);
		assertNotNull(resource);
		resource.load(is, null);
		assertFalse(resource.getContents().isEmpty());
		DocumentRoot root = (DocumentRoot) resource.getContents().get(0);
		EndpointDescriptions edContainer = root.getEndpointDescriptions();
		assertNotNull(edContainer);
		EList<EndpointDescription> endpointDescriptions = edContainer.getEndpointDescription();
		assertEquals(4, endpointDescriptions.size());
		EndpointDescription ed1 = endpointDescriptions.get(0);
		List<Property> properties = ed1.getProperty();
		assertEquals(5, properties.size());

		Map<String, Object> contentMap = new HashMap<>();
		contentMap.put("objectClass", "");
		contentMap.put("osgi.remote.requires.intents", "");
		contentMap.put("testKey", "testValue");
		contentMap.put("endpoint.id", "");
		contentMap.put("service.imported.configs", "org.apache.cxf.ws");

		properties.stream().forEach((p)->{
			if (!contentMap.containsKey(p.getName())) {
				fail("Failed find property with name " + p.getName());
			}
			switch (p.getName()) {
			case "objectClass":
				assertNotNull(p.getArray());
				assertEquals(1, p.getArray().getValue().size());
				//For String base types use XMLTypePackage.Literals.XML_TYPE_DOCUMENT_ROOT__TEXT
				assertEquals("SomeService", p.getArray().getValue().get(0).getMixed().list(XMLTypePackage.Literals.XML_TYPE_DOCUMENT_ROOT__TEXT).get(0));
				break;
			case "osgi.remote.requires.intents":
				assertNotNull(p.getList());
				assertEquals(1, p.getList().getValue().size());
				//For String base types use XMLTypePackage.Literals.XML_TYPE_DOCUMENT_ROOT__TEXT
				assertEquals("confidentiality", p.getList().getValue().get(0).getMixed().list(XMLTypePackage.Literals.XML_TYPE_DOCUMENT_ROOT__TEXT).get(0));
				break;
			case "endpoint.id":
				assertNotNull(p.getMixed());
				//For String base types use XMLTypePackage.Literals.XML_TYPE_DOCUMENT_ROOT__TEXT
				assertEquals("http://somewhere:12345", p.getMixed().list(XMLTypePackage.Literals.XML_TYPE_DOCUMENT_ROOT__TEXT).get(0));
				break;
			default:
				String v = (String) contentMap.get(p.getName());
				if (!v.isEmpty()) {
					assertEquals(v, p.getValue());
				}
				break;
			}

		});
	}

	@Test
	public void testCreateDescription() throws IOException {
		ResourceSet rs = new ResourceSetImpl();
		rs.getPackageRegistry().put(RSAPackage.eNS_URI, RSAPackage.eINSTANCE);
		rs.getResourceFactoryRegistry().getExtensionToFactoryMap().put("xml", new RSAResourceFactoryImpl());
		rs.getResourceFactoryRegistry().getContentTypeToFactoryMap().put(RSAPackage.eCONTENT_TYPE, new RSAResourceFactoryImpl());
		
		String tmp = System.getProperty("user.home");
//		String tmp = System.getProperty("java.io.tmpdir");
		System.out.println("Temp Folder is : " + tmp);
		Resource resource = rs.createResource(URI.createURI(tmp + "/ed.xml"), RSAPackage.eCONTENT_TYPE);
		assertNotNull(resource);
		System.out.println("Resource URI is : " + resource.getURI().toString());
		DocumentRoot root = RSAFactory.eINSTANCE.createDocumentRoot();
		EndpointDescriptions container = RSAFactory.eINSTANCE.createEndpointDescriptions();
		root.setEndpointDescriptions(container);
		
		EndpointDescription ed1 = RSAFactory.eINSTANCE.createEndpointDescription();
		container.getEndpointDescription().add(ed1);
//		EndpointDescription ed2 = RSAFactory.eINSTANCE.createEndpointDescription();
//		container.getEndpointDescription().add(ed2);
//		EndpointDescription ed3 = RSAFactory.eINSTANCE.createEndpointDescription();
//		container.getEndpointDescription().add(ed3);
//		EndpointDescription ed4 = RSAFactory.eINSTANCE.createEndpointDescription();
//		container.getEndpointDescription().add(ed4);
		Property p11 = RSAFactory.eINSTANCE.createProperty();
		p11.setName("objectClass");
		ed1.getProperty().add(p11);
		
		Array array = RSAFactory.eINSTANCE.createArray();
		Value v11 = RSAFactory.eINSTANCE.createValue();
		v11.getMixed().add(XMLTypePackage.Literals.XML_TYPE_DOCUMENT_ROOT__TEXT, "Hallo Welt");
		array.getValue().add(v11);
		p11.setArray(array);
		
		Property p12 = RSAFactory.eINSTANCE.createProperty();
		p12.setName("osgi.remote.requires.intents");
		ed1.getProperty().add(p12);
		
		array = RSAFactory.eINSTANCE.createArray();
		Value v121 = RSAFactory.eINSTANCE.createValue();
		v121.getMixed().add(XMLTypePackage.Literals.XML_TYPE_DOCUMENT_ROOT__TEXT, "confidentiality");
		array.getValue().add(v121);
		Value v122 = RSAFactory.eINSTANCE.createValue();
		v122.getMixed().add(XMLTypePackage.Literals.XML_TYPE_DOCUMENT_ROOT__TEXT, "async");
		array.getValue().add(v122);
		p12.setList(array);
		
		Property p13 = RSAFactory.eINSTANCE.createProperty();
		p13.setName("endpoint.id");
		ed1.getProperty().add(p13);
		p13.getMixed().add(XMLTypePackage.Literals.XML_TYPE_DOCUMENT_ROOT__TEXT, "http://localhost:123");
		
		Property p14 = RSAFactory.eINSTANCE.createProperty();
		p14.setName("service.imported.configs");
		p14.setValue("gecko.rsa");
		p14.setValueType("java.lang.String");
		ed1.getProperty().add(p14);
		
		resource.getContents().add(root);
		resource.save(System.out, null);
	}

}
